import json
import re
from urllib.parse import unquote
from math import fabs
from typing import Sequence

from tracuni.define.const import (
    MASK_PARAMS,
    MASK_PARAMS_PARTIALLY,
    MASK_CHAR,
    MASK_CHAR_PARTIALLY,
    TAG_ESSENTIAL_FIELD_NAMES,
)

MASK_PARAMS_LOW_CASE = {el.lower(): el for el in MASK_PARAMS}
MASK_PARAMS_PARTIALLY_LOW_CASE = {el.lower(): el for el in MASK_PARAMS_PARTIALLY}

essential_target_names = dict(
    (key.lower(), keys[0])
    for keys in [
        key_or_keys if isinstance(key_or_keys, Sequence) and not isinstance(key_or_keys, str) else (key_or_keys,)
        for key_or_keys in TAG_ESSENTIAL_FIELD_NAMES
    ]
    for key in keys
)
essential_keys_to_find = set(essential_target_names.keys())


def mask_partial(row, symbol=MASK_CHAR_PARTIALLY, from_idx=4, to_idx=-4):
    """
    Частичное экранирование строки
    :param row: строка для экранирования
    :param symbol: экранирующий символ
    :param from_idx: стартовая позиция маски
    :param to_idx: конечная позиция маски
    :return:
    """
    row = str(row)
    mask_len = round(len(row) - fabs(from_idx) - fabs(to_idx))
    masked_row = row[:from_idx] + mask_len * symbol + row[to_idx:]
    return masked_row


def mask_full(row, symbol=MASK_CHAR):
    """
    Полное экранирование строки
    :param row: строка
    :param symbol: экранирующий символ
    :return:
    """
    row = str(row)
    masked_row = symbol * len(row)
    return masked_row


def replace_fn(match_obj):
    """
        Пробуем сохранитьб валидный JSON:
            * Если мы число поменяли на строку, оборачиваем в кавычки
            * Если у нас были параметры запроса внутри строки, то возвращаем
            хвостовую кавычку
    :param match_obj:
    :return:
    """
    global collected_essentials
    groups = match_obj.groups()
    start_quote = groups[0]
    keyword = groups[1].lower()
    separator_n_quotes = groups[2] or groups[5]
    value = groups[4] or groups[6]
    value_quotes = groups[3] or ''

    if keyword in MASK_PARAMS_LOW_CASE:
        if keyword in MASK_PARAMS_PARTIALLY_LOW_CASE:
            mask_fn = mask_partial
            keyword = MASK_PARAMS_PARTIALLY_LOW_CASE[keyword]
        else:
            mask_fn = mask_full
            keyword = MASK_PARAMS_LOW_CASE[keyword]
        masked = mask_fn(value)

    else:
        if '\\' in value_quotes:
            value = value[:-2]
        collected_essentials['data.' + essential_target_names[keyword]] = value
        masked = value if value is not None else ''

    return f'{start_quote}{keyword}{separator_n_quotes}{value_quotes}{masked}{value_quotes}'


def prepare_pattern():
    kwords = r'|'.join(MASK_PARAMS | essential_keys_to_find)
    start_quotes = r'[<"&\?]'
    opt_quotes = r'\\?"?'
    opt_space = r'\s*'
    # https://stackoverflow.com/questions/5695240/php-regex-to-ignore-escaped-quotes-within-quotes
    content = r'[^"\\,}}{}]*(?:\\.[^"\\,}}{}]*)*'

    # выцепляет ключи, если они встречаются как ключи в сериализованном JSON:
    #     "{"secret_key": ..., ...}"
    # или если они встречают в JSON, вложенных как экранированные строковые значения:
    #     "{"NON_secret_key": "{\"secret_key\": ...}", ...}"
    # впрочем, так тоже можно:
    #     "{"secret_key": "{\"secret_key\": ...}", ...}"
    content_dict = content.format(*("<",) * 2)
    dict_value = f'({opt_quotes}{opt_space}[:>]{opt_space})({opt_quotes})({content_dict}){opt_quotes}'
    # capture:     ^2=separator & quotes                ^3=quotes     ^4=value

    # выцепляет ключи, если они встречаются как парметры URL запроса внутри JSON значения:
    # {"url_key": "...secret_key=abc&..."}
    content_url = content.format(*("&",)*2)
    url_param = f'({opt_space}={opt_space})({content_url})'
    # capture:    ^5=separator & quotes    ^6=value

    compiled_pattern = re.compile(
        f'({start_quotes})[^/]?({kwords})(?:{dict_value}|{url_param})',
        # ^0=open quote   ^1=keys   ^2-6=see above      : capture
        re.M | re.S | re.I
    )

    return compiled_pattern


pattern = prepare_pattern()
collected_essentials = {}


def do(data):
    global collected_essentials
    collected_essentials = {}
    try:
        data = json.dumps(json.loads(data))
    except (TypeError, json.JSONDecodeError):
        pass
    try:
        data = unquote(data)
    except TypeError:
        pass
    data = pattern.sub(replace_fn, str(data))
    return data, collected_essentials



from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)
from tracuni.schema.pipe_methods import (
    pipe_head,
    pipe_sep_string,
    ext_tracer_path,
    ext_tracer_short_path,
    ext_tracer_full_path
)
from tracuni.misc.helper import json_dumps_decode
from tracuni.schema.shared_rules import rule_inner_out_tracer_context, rule_span_peer_name_from_application_name

def serialize_point_args(x):
    try:
        x = json_dumps_decode(x[0]._asdict()) if x else None
    except Exception as exc:
        return str(exc)
    return x

ruleset = (
    Rule(
        description="Путь по методам, вызвавшим запрос",
        destination=Destination(DestinationSection.REUSE, 'point_path'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Путь по методам, вызвавшим запрос в теги",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.path'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'point_path'),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.method'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_full_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: ("app", "log", x[0]),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')')}),
        ),
        origins=(
            Origin(
                OriginSection.REUSE, 'point_path',
            ),
        ),
        stage=Stage.PRE,
    ),
    rule_span_peer_name_from_application_name,
    Rule(
        description="Вывод аргументов",
        destination=Destination(DestinationSection.SPAN_LOGS, 'args'),
        pipeline=(
            serialize_point_args,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, None),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Разделитель запроса и ответа",
        destination=Destination(
            DestinationSection.SPAN_LOGS,
            'divider',
        ),
        pipeline=(
            lambda _: "--------------------------------",
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                None,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывод результата",
        destination=Destination(DestinationSection.SPAN_LOGS, 'res'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_RESULT, None),
        ),
        stage=Stage.POST,
    ),
)



import datetime
import decimal
import inspect
import json
import re
import os
import sys
from itertools import islice
from typing import Iterable
import logging

import tracuni.define.errors as err

try:
    from aiohttp.web_request import Request
    from multidict import CIMultiDictProxy, CIMultiDict
except ImportError:
    Request = None
    CIMultiDict = None
    CIMultiDictProxy = None


try:
    # noinspection PyPackageRequirements
    from psycopg2 import ProgrammingError as PostgreSQLError
except ImportError:
    class PostgreSQLError(BaseException):
        pass

try:
    # noinspection PyPackageRequirements
    from asyncpg.protocol.protocol import Record as PostgreSQLRecord
except ImportError:
    PostgreSQLRecord = None


pattern_to_nerf_tail_digits = re.compile(r"(/(\d*)*?)*?$")


def is_iter_but_str(item):
    return not isinstance(item, str) and isinstance(item, Iterable)


def to_iter(item):
    if is_iter_but_str(item):
        return item
    else:
        return item,


def _json_encoder(obj):
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S.%f%z')
    if isinstance(obj, datetime.date):
        return obj.strftime('%Y-%m-%d')
    if isinstance(obj, datetime.time):
        return obj.strftime('%H:%M:%S.%f%z')
    if isinstance(obj, datetime.timedelta):
        return obj.total_seconds()
    if isinstance(obj, bytes):
        try:
            return obj.decode('UTF8')
        except Exception:
            return str(obj)
    return repr(obj)


def json_dumps_decode(data, indent=0):
    return json.dumps(data, default=_json_encoder, indent=indent)


def get_project_path():
    return os.path.abspath(os.path.dirname(sys.argv[0]))


relative_path = __file__.replace(get_project_path(), '')
project_path_pattern = re.compile(
    r'^{}/'.format(
        get_project_path(),
    )
)


def compose_own_call_stack(
        skip_methods,
        skip_paths,
        result_length=1,
        look_up_stop_idx=27,
        from_index=0,
        frame_list=None,
):
    outer_frames = frame_list if frame_list is not None else inspect.stack()

    def is_project_outer_frame(item):
        try:
            file_name = item.f_code.co_filename
            return (
                project_path_pattern.match(file_name)
                and
                "site-packages" not in file_name
                and
                "lib/python" not in file_name
            )
        except (IndexError, AttributeError):
            return False

    def compose_frame_path_object(item):
        try:
            file_name = project_path_pattern.sub(
                '',
                item.f_code.co_filename,
                1,
            )
            method_name = item.f_code.co_name
            method_string_no = str(item.f_code.co_firstlineno)
        except (IndexError, AttributeError):
            return None
        if method_name in skip_methods:
            return None
        if skip_paths in file_name:
            return None
        return {
            'full_path': file_name + ':' + method_string_no + ':' + method_name,
            'short_path': method_name,
        }

    lookup_list = [
        el
        for el in outer_frames[from_index:look_up_stop_idx]
        if is_project_outer_frame(el)
    ]
    composed_data_list = [
        compose_frame_path_object(el)
        for el in lookup_list
    ]
    compacted_list = [
        el
        for el in composed_data_list
        if el is not None
    ]
    result_dicts = list(islice(compacted_list, result_length))

    return result_dicts


def dict_from_json(data):
    """
    Return dict from json-string or dict
    :param data: string or dict
    :return: dict
    """
    try:
        data_dict = json.loads(data)
    except ValueError as e:
        logging.warning('JSON parse error: %s Data: %s.', e, data)
        data_dict = data
    return data_dict


def modify_aiohttp_request_headers(
    arg_tuple,
    request_position,
    new_headers,
):
    if (
        Request
        and
        len(arg_tuple) > 1
        and
        isinstance(arg_tuple[1], Request)
    ):
        request = arg_tuple[1]
        new_headers = CIMultiDictProxy(
            CIMultiDict({
                **dict(request.headers),
                **new_headers,
            })
        )
        request = request.clone(
            headers=new_headers
        )
        return (
            *arg_tuple[0:request_position],
            request,
            *arg_tuple[request_position + 1:]
        )
    else:
        return arg_tuple


def log_tracer_error(exc, debug_is_on):
    if debug_is_on:
        if not isinstance(
            exc,
            (err.HTTPPathSkip, err.SpanNoParentException)
        ):
            logging.exception(exc)
        else:
            logging.debug(exc)


def nerf_tail_digits(source_string):
    return pattern_to_nerf_tail_digits.sub("", source_string, 1)



from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)
from tracuni.schema.pipe_methods import (
    pipe_head,
    pipe_sep_string,
)
from tracuni.schema.shared_rules import (
    rule_inner_in_tracer_context,
    rule_span_peer_name_from_application_name,
)

ruleset = (
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: ("app", "retry", x[0]),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')')}),
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, 'retries',
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Занести максимальное количество попыток в теги",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'app.retries.max'
        ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, 'retries',
            ),
        ),
        stage=Stage.PRE,
    ),
    rule_inner_in_tracer_context,
    rule_span_peer_name_from_application_name,
)



from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)
from tracuni.schema.pipe_methods import (
    pipe_head,
)
from tracuni.schema.shared_rules import rule_amqp_out_tracer_context, rule_span_peer_name_from_application_name

ruleset = (
    rule_amqp_out_tracer_context,
    Rule(
        description="Наименование участка повторного запуска задачи",
        destination=Destination(DestinationSection.SPAN_NAME, 'name'),
        pipeline=(lambda _: 'amqp::retry',),
        origins=(Origin(OriginSection.ADAPTER, None),),
        stage=Stage.PRE,
    ),
    rule_span_peer_name_from_application_name,
    Rule(
        description="Вывести повторы задачи в теги",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.retry'),
        pipeline=(
            lambda task_data: f"{task_data[0]}/{task_data[1]}"
            if task_data else None,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, "retry",
            ),
            Origin(
                OriginSection.POINT_ARGS, "retry_count",
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести идентификатор задачи в теги",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.task'),
        pipeline=(
            pipe_head,
            lambda tid: f"#{tid}" if tid else None,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, "tid",
            ),
        ),
        stage=Stage.PRE,
    ),
)



from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
    PipeCommand,
)
from tracuni.schema.pipe_methods import (
    ext_sql_args,
    ext_tracer_out_point,
    ext_tracer_path,
    ext_tracer_full_path,
    pipe_head,
    pipe_sep_string,
    pipe_db_service_name_from_config,
    pipe_catch_essential,
    pipe_mask_secret_catch_essentials,
    pipe_dump,
    pipe_cut,
    ext_api_kind,
    ext_sql_response,
)


ruleset = (
    Rule(
        description="Путь по методам, вызвавшим запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.path'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.method'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_full_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            lambda x: ('db', *x),
            partial(pipe_sep_string, **{'sep': ("::", " (", ")")}),
        ),
        origins=(
            Origin(
                OriginSection.CALL_STACK,
                partial(ext_tracer_out_point, **{'shift': 0}),
            ),
            Origin(
                OriginSection.CALL_STACK,
                ext_tracer_out_point,
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Наименование партнерского сервиса сохранить для "
                    "переиспользования",
        destination=Destination(
            DestinationSection.REUSE,
            'peer_name',
          ),
        pipeline=(
            pipe_head,
            pipe_db_service_name_from_config,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, 'self'),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Наименование партнерского сервиса записать в участок",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
          ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'peer_name'),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести праметры SQL запроса в аннотации",
        destination=Destination(DestinationSection.SPAN_LOGS, 'SQL_params'),
        pipeline=(
            pipe_head,
            # pipe_catch_essential,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS,
                ext_sql_args,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести ответ сервера SQL",
        destination=Destination(DestinationSection.SPAN_LOGS, 'SQL_response'),
        pipeline=(
            pipe_head,
            # pipe_catch_essential,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_RESULT,
                ext_sql_response,
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Запомнить ошибку для вывода",
        destination=Destination(DestinationSection.REUSE, 'error'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.SPAN, 'error'),
        ),
        stage=Stage.POST,
    ),
)



import aiozipkin.transport as azt
import logging
from urllib.parse import quote
from json import loads, JSONDecodeError

from tracuni.side_steps.stat.adapter import StatsD
from tracuni.misc.select_coroutine import get_coroutine_decorator
from tracuni.define.const import STATSD_DURATION_FACTOR
async_decorator = get_coroutine_decorator()


class TracerTransport(azt.Transport):
    # noinspection PyProtectedMember
    def __init__(
            self,
            logging_conf,
            statsd_conf,
            loop,
            use_tornado=False,
    ):
        if use_tornado:
            raise Exception("Tornado is not supported in this vwrsion")
        tracer_url = logging_conf.url
        send_interval = logging_conf.send_interval
        super(TracerTransport, self).__init__(
            tracer_url,
            send_interval=send_interval,
            loop=loop
        )
        self.loop = loop
        self.__tracer = logging_conf.name
        self.use_tornado = use_tornado

        statsd_conf = StatsD.read_statsd_configuration(statsd_conf)
        self.stats = None
        if statsd_conf.enable:
            self.stats = StatsD(statsd_conf._asdict())
            self.stats.connect()
        else:
            logging.warning("StatsD send has NOT been started")
        self.statsd_is_configured = statsd_conf.enable and self.stats
        if self.statsd_is_configured:
            logging.info("StatsD send is nominal")

    @async_decorator
    def _send(self, record):
        try:
            if self.statsd_is_configured:
                yield self._send_to_statsd(record)
        except Exception as e:
            logging.exception(e)

        try:
            if self.__tracer == 'zipkin':
                yield super(TracerTransport, self).send(record)
            else:
                self._queue = []
        except Exception as e:
            logging.exception(e)

    def send(self, record):
        self.loop.create_task(self._send(record))

    @async_decorator
    def _send_to_statsd(self, rec):
        if self.stats:
            name = rec._name
            tags = rec._tags or {}

            api_kind = name.split('::')[0]
            span_side = rec._kind
            json_method = tags.get('data.method')
            json_response_code = str(tags.get('data.code'))
            widget_id = tags.get('app.widget')
            app_method = tags.get('app.method', '').split(':')[-1]
            rq_hostname = tags.get('rq.url.hostname')

            try:
                rsp_status = int(tags.get('rsp.status') or 0)
            except ValueError:
                rsp_status = 0
            duration = int(round(rec._duration / STATSD_DURATION_FACTOR))

            span_side_map = {
                'SERVER': 'in',
                'CLIENT': 'out',
            }

            if tags.get('error') is not None:
                own = tags.get('error.own')
                xls = tags.get('error.class')

                if own or api_kind == '<unknown>':
                    variant = "tracer_own"
                else:
                    variant = f"{api_kind}_{span_side_map.get(span_side)}"

                # ALL:ALL if error
                # counter <unit_name>.exception,class=x,variant=x
                # class = exception class name
                # variant = ((HTTP|AMQP|DB_IN|OUT)|tracer_own)
                self.stats.exception_count_glork(xls, variant)

            if api_kind == 'http':
                if span_side == 'CLIENT':
                    try:
                        tm_direction = loads(rec._annotations[1].value)['X-TM-DIRECTION']
                    except (AttributeError, IndexError, KeyError, JSONDecodeError):
                        tm_direction = 'notm'


                    # HTTP:OUT
                    # timer <unit_name>.http_requests,code=x,hostname=x
                    # code = peer http response code
                    # hostname = peer address
                    self.stats.http_out_time_glork(
                        rsp_status, rq_hostname, tm_direction, json_response_code,
                        duration
                    )
                elif span_side == 'SERVER':
                    # HTTP:IN
                    # counter <unit_name>.gen_response_code,code=x,widget_id=x,method=x
                    # code = API response code, where 0 = success
                    # widget_id = widget data from request
                    # method = API method from request
                    if json_response_code is not None:
                        self.stats.http_in_code_count_glork(
                            json_response_code, widget_id, json_method
                        )
                    # HTTP:IN
                    # timer <unit_name>.handler_timers,name=x,widget_id=x,method=x
                    # name = handler method name
                    # widget_id = widget data from request
                    # method = API method from request
                    self.stats.http_in_method_time_glork(
                        app_method, widget_id, json_method,
                        duration,
                    )

            if api_kind == 'db':
                # DB:OUT
                # timer <unit_name>.db_query_function,func_name=<method_name>
                # func_name = db api method name
                self.stats.db_time_glork(
                    app_method,
                    duration,
                )
                self.stats.db_pool_free_count_glork(tags.get('app.db_pool_free_slots'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `trac_uni` package."""

import pytest

from click.testing import CliRunner
from tracuni.trac_uni import main


# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


@pytest.fixture
def argument():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    from tracuni.trac_uni import cube_primitive_test
    value = 2
    return value, cube_primitive_test


def test_content(argument):
    """Sample pytest test function with the pytest fixture as an argument."""
    value, fn = argument
    fn(value)


@pytest.fixture
def runner():
    return CliRunner()


def test_main(runner):
    result = runner.invoke(main)

    assert result.exit_code == 0
    assert 'Demo app for trac uni package' in result.output
    assert '2 x 2 = 4\n' in result.output

    help_result = runner.invoke(main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Тестовые типы и экземпляры обрабтчика схем
"""

import sys
import pytest


@pytest.fixture
def engine_types(mocker, mock_standard_schema, stage_to_sections):
    """Импорт мета-класса, осуществляющего кеширование экземпляров процессоров
    схем
    """
    if 'trac_uni.schema.standard' in sys.modules:
        del sys.modules['trac_uni.schema.standard']
    if 'trac_uni.schema.engine.engine' in sys.modules:
        del sys.modules['trac_uni.schema.engine.engine']
    mocker.patch(
        'typing.TYPE_CHECKING',
        True
    )

    # noinspection PyUnresolvedReferences
    from tracuni import trac_uni

    mocker.patch(
        'trac_uni.schema.standard.standard_schema',
        mock_standard_schema
    )
    mocker.patch(
        'trac_uni.schema.standard.stage_to_sections',
        stage_to_sections
    )
    mocker.patch(
        'trac_uni.schema.standard.secret_mask_method_name',
        'test_pipe_secret'
    )
    mocker.patch(
        'trac_uni.schema.standard.secret_mask_method_name',
        'test_pipe_secret'
    )

    from tracuni.trac_uni import SchemaEngineCache  # noqa
    from tracuni.trac_uni import SchemaEngine
    from tracuni.trac_uni import SchemaEngineFeed
    from tracuni import trac_uni as err

    return SchemaEngine, SchemaEngineCache, SchemaEngineFeed, err


@pytest.fixture
def simple_engine(engine_types, variants):
    return engine_types[0](variants[0])


@pytest.fixture
def both_custom_rules_engine(engine_types, variants, mock_standard_schema):
    return engine_types[0](
        variants[0],
        point_ruleset=mock_standard_schema[variants[0]],
        custom_schema=mock_standard_schema,
    )



import logging
import aiozipkin as az
import asyncio
import asyncpg

from app.core.app import Component
from app.core.helper import PrepareError


def db_decorator(func):
    async def wrapper(*args, **kwargs):
        res = await func(*args, **kwargs)
        return res
    return wrapper


class DB(Component):
    def __init__(self, config):
        super(DB, self).__init__()
        dsn = 'postgres://%s:%s@%s:%s/%s' % (
            config['username'],
            config['password'],
            config['host'],
            config['port'],
            config['dbname']
        )
        self.connect_max_attempts = config.get('connect_max_attempts', 10)
        self.connect_retry_delay = config.get('connect_retry_delay', 5)
        self._dsn = dsn
        self.pool = None

    async def _connect(self):
        while True:
            try:
                self.pool = await asyncpg.create_pool(
                    dsn=self._dsn,
                    max_size=self.app.config['system']['pool_max_size'],
                    min_size=self.app.config['system']['pool_min_size'],
                    max_queries=self.app.config['system']['pool_max_queries'],
                    max_inactive_connection_lifetime=(
                        self.app.config['system']['pool_max_inactive_connection_lifetime']
                    ),
                )
            except Exception as e:
                await asyncio.sleep(5)
                logging.error('DB connect error: %s %s' % (self.__class__.__name__, str(e)))
                logging.error('DSN: %s' % self._dsn)
                logging.error('Trying to reconnect to database...')
            else:
                break

    @db_decorator
    async def execute(self, context_span, db_id, sql, *args):
        async with self.pool.acquire() as conn:
            try:
                await conn.execute(sql, *args)
            except Exception as e:
                logging.error('DB _execute %s %s error: %s' % (sql, self.__class__.__name__, str(e)))
                res = e
            else:
                res = None
        return context_span, db_id, res

    @db_decorator
    async def query(self, context_span, db_id, sql, *args):
        async with self.pool.acquire() as conn:
            try:
                res = await conn.fetchrow(sql, *args)
            except Exception as e:
                logging.error('DB _query %s %s error: %s' % (sql, self.__class__.__name__, str(e)))
                res = None
        return context_span, db_id, res

    @db_decorator
    async def query_all(self, sql, *args):
        async with self.pool.acquire() as conn:
            try:
                res = await conn.fetch(sql, *args)
            except Exception as e:
                logging.error('DB _query_all %s %s error: %s' % (sql, self.__class__.__name__, str(e)))
                res = None
        return res

    async def prepare(self):
        self.app.log_info('Connecting to %s' % self._dsn)
        for i in range(self.connect_max_attempts):
            try:
                await self._connect()
                self.app.log_info('Connected to %s' % self._dsn)
                return
            except Exception as e:
                self.app.log_err(str(e))
                await asyncio.sleep(self.connect_retry_delay)
        raise PrepareError('Could not connect to %s' % self._dsn)

    async def start(self):
        pass

    async def stop(self):
        self.app.log_info('Disconnecting from %s' % self._dsn)
        await asyncio.sleep(0.3)

    async def ping(self):
        async with self.pool.acquire() as conn:
            try:
                await conn.execute('select 1')
            except Exception as e:
                logging.error('DB _execute ping')
                res = e
            else:
                res = None
        return res



import re
import json
import io
from collections import namedtuple
from itertools import islice
from urllib.parse import parse_qs, urlparse
from json import JSONDecodeError

from aiohttp import BytesPayload
from aiohttp.web_response import Response

from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    Dict,
    Sequence,
    Tuple,
)

from tracuni.schema.lib import (
    catch_essential,
    mask_secret,
    make_cut,
    pick_items,
)
from tracuni.misc.helper import (
    PostgreSQLRecord,
    json_dumps_decode,
)
from tracuni.define.const import (
    CUT_LIMIT,
    CUT_LINE,
    UNKNOWN_NAME,
    LOG_STRING_WRAP,
)
from tracuni.define.type import (
    TeeDescriptor,
    PipelineTee,
    DestinationSection,
    PipeCommand,
)
from tracuni.misc.select_coroutine import get_coroutine_decorator
async_decorator = get_coroutine_decorator()

if TYPE_CHECKING:
    from tracuni.adapter import AdapterType


pattern_module_method_names = re.compile(r'.*/(.*)\.py:\d+\.(.*)$')


###############################################################################
# Новые проверенные


def pipe_debug(data):
    return PipeCommand.DEBUG


@async_decorator
def ext_http_in_aiohttp_request(arguments: namedtuple) -> Dict[str, Any]:
    """Получить данные входящего HTTP запроса для AIOHTTP

    Parameters
    ----------
    arguments
        Аргументы точки вызова

    Returns
    -------
        Описание и нагрузку запроса, включая заголовки трассера

    """
    request = None
    widget_id = None
    if hasattr(arguments, 'self'):
        request = getattr(arguments.self, 'request', None)
        system = getattr(arguments.self, 'system', None)
        if system:
            widget_id = system.widget_id
        if widget_id is None:
            widget_id = getattr(arguments.self, 'widget_id', None)
    if not request:
        request = getattr(arguments, 'request', None)

    if request is None:
        return {}

    # body = request._payload._buffer
    body = getattr(getattr(request, "_payload", None), "_buffer",  None)
    if body is not None:
        if len(body):
            body = body[0].decode()
        else:
            body = (getattr(request, "_read_bytes", b'{}') or b'{}').decode()
        try:
            body = json.loads(body)
        except Exception:
            body = pipe_qs_to_dict(body)
    else:
        try:
            body = yield from request.json()
        except JSONDecodeError:
            body = yield from request.text()
            body = pipe_qs_to_dict(body)

    query = None
    if hasattr(request, 'query'):
        query = dict(request.query)

    headers = {}
    zipkin_headers = {}

    content_length = request.content_length

    marks = {
        "redr": "REDIRECT",
        "ajax": "AJAX",
    }

    mark = ""
    proc_id = body.get("proc_id", body.get("form_request_id"))
    if proc_id:
        proc_id += ":ajax"
    zipkin_ids = getattr(request, "query", {}).get("X-B3", request.cookies.get("X-B3", proc_id))
    if zipkin_ids:
        trace_id, span_id, *link_type = zipkin_ids.split(":")
        if link_type:
            link_type = link_type[0]
        else:
            link_type = span_id
            span_id = None
        mark = marks.get(link_type, mark)
        zipkin_headers = {
            "X-B3-TraceId": trace_id,
            "X-B3-SpanId": span_id,
            "X-B3-Flags": '0',
            "X-B3-Sampled": '1',
        }
    else:
        for header_name, header_value in request.headers.items():
            header_name_up = header_name.upper()
            if header_name_up.startswith('X-B3-'):
                zipkin_headers[header_name_up] = header_value
            else:
                headers[header_name] = header_value

    url_path = request.url.path
    url_wo_query = str(request.url.parent) + request.url.path
    remote_ip = request.remote

    body.pop('jsonParams', None)
    res = {
        'body': body,
        'query_str': '?' + request.query_string if request.query_string else '',
        'query_dict': query,
        'url': url_wo_query,
        'url_path': url_path,
        'url_host': request.url.parent,
        'remote_ip': remote_ip,
        'method': request.method,
        'widget_id': widget_id,
        'length': content_length,
        'headers': headers,
        'zipkin_headers': zipkin_headers,
        'mark': mark,
    }
    return res


def ext_http_in_tornado_request(arguments: namedtuple) -> Dict[str, Any]:
    """Получить данные входящего HTTP запроса для Tornado

    Parameters
    ----------
    arguments
        Аргументы точки вызова

    Returns
    -------
        Описание и нагрузку запроса, включая заголовки трассера

    """
    widget_id = None
    request = None
    if hasattr(arguments, 'self'):
        request = getattr(arguments.self, 'request', None)
        system = getattr(arguments.self, 'system', None)
        if system:
            widget_id = getattr(system, 'widget_id', None)
        if widget_id is None:
            widget_id = getattr(arguments.self, 'widget_id', None)

    if not request:
        request = getattr(arguments, 'request', None)
    if request is None:
        return {}

    body_str = request.body.decode() or "{}"
    try:
        body = json.loads(body_str)
    except JSONDecodeError:
        body = body_str

    zipkin_headers = {}

    headers = dict(request.headers)
    content_length = headers.get('Content-Length', len(body_str))
    for header_name, header_value in request.headers.items():
        if header_name.startswith('X-B3-'):
            zipkin_headers[header_name] = header_value
        else:
            headers[header_name] = header_value

    url = request.protocol + "://" + request.host + request.path
    url_path = request.path

    remote_ip = request.remote_ip

    try:
        if 'pop' in body:
            body.pop('jsonParams'
                     ''
                     '', None)
    except TypeError:
        pass

    res = {
        'body': body,
        'query_str': '?' + request.query if request.query else '',
        'query_dict': parse_qs(request.query or ''),
        'url': url,
        'url_path': url_path,
        'url_host': request.host,
        'remote_ip': remote_ip,
        'method': request.method,
        'widget_id': widget_id,
        'length': content_length,
        'headers': headers,
        'zipkin_headers': zipkin_headers,
    }
    return res


def ext_out_headers(args):
    self = getattr(args, 'self', None)
    if self is not None:
        return getattr(getattr(self, 'request', None), 'headers', None)
    return getattr(args, 'headers', None)


def pipe_head(seq: Sequence[Any], indexes=0) -> Any:
    """Получить первый элемент из обрабатываемого набора данных

    Parameters
    ----------
    seq
        Список извлечённых данных (каждый источник представлен одним элементом)

    Returns
    -------
        Первый элемент списка
    """
    v = pick_items.do(seq, indexes)
    return v


def pipe_check_skip(data: Tuple[str, 'AdapterType']):
    """Для входящих HTTP: находится ли адрес обработчика в списке игнорируемых

    Parameters
    ----------
    data
        кортеж: проверяемый путь и ссылка на адаптер

    Returns
    -------
        флаг, показывающий надо ли пропускать
    """
    path, adapter = data
    flag = adapter.should_not_trace(path)
    return flag


def pipe_sep_string(
    data: Sequence,
    sep: Union[Sequence[str], str] = ' '
) -> str:
    """Преобразует несколько значений в одну строку с разделителями

        Можно передавать несколько разделителей, тогда первый станет
        разделителем первой пары значений, второй - второй и так далее,
        если раздлителей меньше. чем пар значений, то последний разделитель
        будет использован для всех оставшихся

    Parameters
    ----------
    data
        Список обрабатываемых данных
    sep
        Набор разделителей для объединения списка

    Returns
    -------
        Строку значений с указанными разделителями
    """
    if not data:
        return ''

    if isinstance(sep, str):
        sep = (sep, )

    if not sep:
        sep = ('', )

    placeholder = '{}'
    place_len = len(placeholder)
    n = len(data)
    tmpl = placeholder * n
    sep_tail = (sep[-1],) * (n - len(sep))

    placeholder_line = [
        tmpl[idx:idx + place_len]
        for idx in range(0, len(tmpl), place_len)
    ]
    separators = sep + sep_tail
    zipped = list(zip(placeholder_line, separators))
    if len(data) != len(sep):
        zipped[-1] = (zipped[-1][0],)

    composed = ''.join([
        el
        for sublist in zipped
        for el in sublist
    ]).format(*data)
    return composed


def pipe_catch_essential(data):
    # return catch_essential.do(data)
    # obsolete - essentials are caught inside mask secret method
    return data


def pipe_mask_secret_catch_essentials(data):
    return mask_secret.do(data)


def pipe_by_name(data, name: str) -> Any:
    return getattr(
        data,
        name,
        data.get(name)
        if getattr(data, "get", False)
        else UNKNOWN_NAME
    )


def pipe_complex_result(data, key):
    if (
        isinstance(data, tuple)
        and
        len(data) == 2
        and isinstance(data[0], Response)
    ):
        data = data[0]
    if isinstance(data, Response):
        value = getattr(data, key, None)
        if isinstance(value, bytes):
            value = value.decode()
        return value
    if not isinstance(data, dict) or not key:
        return data
    res = data.get(key, {})
    if key in ('response_headers', 'request_headers'):
        res = dict((k, v) for k, v in res.items() if not k.startswith('X-B3'))
    return res


def pipe_dump(data):
    if isinstance(data, bytes):
        return data.decode()
    if isinstance(data, (dict, tuple, list)):
        return json_dumps_decode(data, indent=4)
    return data


def pipe_cut(data, cut_limit=CUT_LIMIT, cut_line=CUT_LINE):
    if isinstance(data, (dict, tuple, list)):
        prepared_data = json_dumps_decode(data, indent=4)
    else:
        prepared_data = str(data)
    cut, _, is_cut = make_cut.do(
        prepared_data,
        cut_limit,
        cut_line
    )
    if len(cut) > 2:
        cut = cut[1:-1].replace('\\\"', '"').replace('\\n', '\n')
    return cut if is_cut else data


def pipe_inject_headers(data, prefix_key=None):
    if len(data) < 2:
        data = (*data, {})
    if len(data) < 3:
        data = (*data, UNKNOWN_NAME)
    span, headers, service_name = data
    if headers is None:
        headers = {}

    headers = span.enrich_headers(headers, prefix_key)
    name_record = {'X-B3-ServiceName': service_name}
    if prefix_key:
        headers[prefix_key].update(name_record)
    else:
        headers.update(name_record)
    return headers


def ext_tracer_out_point(track, name_variant='short', shift=1):
    data_l = len(track)

    name_variant += '_path'
    if name_variant == 'compact_path':
        full_name = track[shift]['full_path']
        compact_name = pattern_module_method_names.match(full_name).groups()
        track[shift][name_variant] = '.'.join(compact_name)

    if data_l > shift:
        return track[shift][name_variant]
    elif data_l:
        return track[0][name_variant]
    return ''


def ext_http_out_url(params):
    res = getattr(params, 'url', '')
    res = ''.join([urlparse(res)[0], '://', *urlparse(res)[1:3]])
    return res


def pipe_db_service_name_from_config(data):
    """Подставляет имя партнерского сервиса из конфигурации, если данный URL
    в ней есть, иначе оставляет url

    Parameters
    ----------
    data type: tuple
        (словарь настроек, строка url)

    Returns
    -------
        Наименование партнерского сервиса
    """
    return (getattr(data, "config", {}) or {}).get('dbname', 'db')


def pipe_service_name_from_config(data):
    """Подставляет имя партнерского сервиса из конфигурации, если данный URL
    в ней есть, иначе оставляет url

    Parameters
    ----------
    data type: tuple
        (словарь настроек, строка url)

    Returns
    -------
        Наименование партнерского сервиса
    """
    val = data[1]
    if len(data) > 1:
        try:
            val = data[0][data[1]].replace('_url', '')
        except (IndexError, KeyError):
            pass
    return val


def ext_send_request_args(data):
    v = getattr(data, 'req', None)
    if not v:
        v = getattr(data, 'body', '')
    if isinstance(v, str):
        v_dict = parse_qs(v)
        if not v_dict:
            step = LOG_STRING_WRAP
            chunks = [v[idx:idx+step] for idx in range(0, len(v), step)]
            v = '\n'.join(chunks)
        else:
            v = v_dict
    return v


def ext_api_kind(engine):
    return engine.schema_feed.variant.api_kind.name.lower()


def ext_tracer_path(track, length=5):
    track_short_list = tuple(islice(
        (
            path_element
            for path_element in (el.get('short_path') for el in track)
            if path_element not in ('run', 'run_loop', 'main')
        ),
        0,
        length
    ))
    return (' > '.join(
        ('{}',)*len(track_short_list)
    )).format(*reversed(track_short_list))


def ext_tracer_full_path(track):
    full_path = ''
    if len(track) > 1:
        full_path = track[1].get('full_path')
    return full_path


def ext_tracer_short_path(track):
    full_path = ''
    if len(track) > 1:
        full_path = track[1].get('short_path')
    return full_path


def ext_sql_args(arguments):
    # noinspection PyProtectedMember
    res = dict(
        (k, v)
        for k, v in arguments._asdict().items()
        if k not in ('self', 'sql')
    )
    if 'method_arguments_without_names' in res and len(res) == 1:
        res = res['method_arguments_without_names']
        if res and len(res) == 1:
            res = res[0]
    return res


def ext_sql_response(res):
    if PostgreSQLRecord and isinstance(res, PostgreSQLRecord):
        return dict(res)
    return res


###############################################################################
# Старые


# noinspection SpellCheckingInspection
def pipe_convert_dict(dikt: dict):
    return convert(dikt)


def pipe_convert_bytes(data):
    res = annotate_bytes(data)
    if res == 'None':
        res = None
    return res


def get_space(_):
    return ' '


def pipe_identity(v):
    return v


def pipe_qs_to_dict(data):
    if isinstance(data, str):
        res = parse_qs(data)
        res = dict((k, v[0] if len(v) == 1 else v) for k, v in res.items())
        return res
    return data


def ext_http_in_path(arguments):
    """Поулчить адрес входящего запроса

    Parameters
    ----------
    arguments
        Аргументы точки вызова

    Returns
    -------
        URL из запроса, если найден в ожидаемых атрибутах
    """
    request = getattr(arguments.self, 'request', None)
    if not request:
        request = getattr(arguments, 'request', None)
    return request.path


def load_dict(data):
    try:
        return json.loads(str(data))
    except JSONDecodeError:
        string_payload_value = getattr(data, '_value', None)
        if string_payload_value:
            return string_payload_value
        return str(data)


def convert(data):
    data_type = type(data)

    if data_type == bytes:
        return data.decode()
    if data_type in (str, int, bool):
        return str(data)
    if data is None:
        return None

    if data_type == dict:
        data = data.items()

    return data_type(map(convert, data))


def annotate_bytes(data, fallback_value='<unknown>'):
    if isinstance(data, BytesPayload):
        pl = io.BytesIO()
        # noinspection PyTypeChecker
        data.write(pl)
        data = pl.getvalue()
    try:
        data_str = data.decode("UTF8")
    except (AttributeError, UnicodeDecodeError):
        data_str = str(data)

    return data_str or fallback_value


def pipe_normalize_json_string(json_data):
    try:
        return json.dumps(json.loads(json_data))
    except (TypeError, json.JSONDecodeError):
        return json_data


def pipe_get_send_request_core(point_args):
    url = getattr(point_args, 'url', None)
    parsed_url = urlparse(url)
    url_hostname = parsed_url.hostname
    url_address = ''.join([parsed_url[0], '://', *parsed_url[1]])
    url_path = parsed_url[2] or '/'
    query_parsed = parsed_url[4]
    body = None if query_parsed else getattr(point_args, 'body', None)
    return {
        "body": body,
        "query": query_parsed,
        "url": url,
        "url_hostname": url_hostname,
        "url_address": url_address,
        "url_path": url_path,
        "method": getattr(point_args, 'method', 'POST'),
        "headers": getattr(point_args, 'headers', None),
    }


def pipe_get_send_request_tornado(point_args):
    url = getattr(point_args, 'url', None)
    parsed_url = urlparse(url)
    url_hostname = parsed_url.hostname
    url_address = ''.join([parsed_url[0], '://', *urlparse(url)[1]])
    url_path = parsed_url[2] or '/'
    body_var_a = getattr(point_args, 'req', None)
    body_var_b = getattr(point_args, 'request_arr', None)
    query = getattr(point_args, 'body', None)
    return {
        "body": body_var_a or body_var_b,
        "query": query,
        "url": url,
        "url_hostname": url_hostname,
        "url_address": url_address,
        "url_path": url_path,
        "method": getattr(point_args, 'method', 'POST'),
        "headers": getattr(point_args, 'headers', None),
    }





def do(
    stringified_response: str,
    cut_limit: int = 4000,
    cut_line: str = '<...>'
):
    try:
        half_cut_limit = abs(int(cut_limit / 2))
    except TypeError:
        half_cut_limit = 2000
    was_cut = False
    stringified_response_len = len(stringified_response)
    if stringified_response_len > half_cut_limit * 2:
        was_cut = True
        stringified_response = '"{head}\\n{cut}\\n{tail}"'.format(**{
            'head': stringified_response[:half_cut_limit],
            'cut': cut_line.format(**{
                'cut_limit': cut_limit,
                'ttl_len': stringified_response_len
            }),
            'tail': stringified_response[-half_cut_limit:],
        })

    return stringified_response, stringified_response_len, was_cut



import json
from textwrap import dedent as ____
from tracuni.define.type import PipeCommand
import logging


def debug_output(point_context, debug_info):
    try:
        point_name = point_context.point_ref.__qualname__
        # noinspection PyProtectedMember
        adapter_config = json.dumps(point_context.adapter.config._asdict(), indent=2)
        variant = point_context.engine.schema_feed.variant
        schema = json.dumps(dict(
            (k.name, "{}::{}".format(rule.destination.section.name, rule.destination.name))
            for k, v in point_context.engine.extractors.items()
            for rule in v
        ))
        value = debug_info.get('value')
        if not value:
            return
        processor_idx = debug_info['processor_idx']
        extractor = debug_info['extractor']
        extractor_description = extractor.description
        processor = debug_info['processor']
        if processor_idx > 0:
            processor = extractor.pipeline[processor_idx - 1]
            if processor_idx >  1 and processor == PipeCommand.TEE:
                processor = extractor.pipeline[processor_idx - 2]
        processor_name = processor.__name__

        output = ____(f"""


        Calling app method <<{point_name}>> marked as: <<{variant}>>,
        processing with <<{processor_name}>> method from rule described as: <<{extractor_description}>>
        
        Values received:
        ================
        {value}
        ================
        """)

        logging.info(output)
        return output

    except Exception as exc:
        logging.debug("Pipeline debug ruined itself")
        logging.exception(exc)



class TracerCollectedException(Exception):
    """Ошибка в приложении, перехватывается трасером"""
    def __init__(self, result=None, exc=None):
        self.result = result
        self.exc = exc


class TracerCollectedHTTPException(TracerCollectedException):
    """Ошибка в приложении, код ошибки протокола в HTTP заголовке, перехватывается трасером"""
    def __init__(self, msg, error_code):
        self.error_code = error_code
        self.msg = msg
        super().__init__(self, msg)

    def __str__(self):
        return repr(self.msg)


class TracerCollectedApplicationException(TracerCollectedException):
    """Ошибка в приложении, код ошибки приложения в JSON, перехватывается трасером"""


class ProcessorInitNoneOrWrongTypeVariant(Exception):
    """Variant is not found in provided VariantSchemaMaster
    dependencies"""


class ProcessorInitWrongTypeCustomSchema(Exception):
    """Provided Custom Schema is of wrong type"""


class ProcessorExtractCantSetDestinationAttr(Exception):
    """Invalid destination"""


class NoSpanException(Exception):
    """Attempt to process span without span itself, skipping"""


class SpanIsNotStartedException(Exception):
    """Attempt to fill in or finish span that has not been started"""


class SpanNoParentException(Exception):
    """Outgoing point without parent span,
        it's no error just a normal case flow
    """


class SpanNoStageForMethod(Exception):
    """Method decorated as stage-bound has no stage linked
    """


class HTTPPathSkip(Exception):
    """Skip HTTP path, it's no error just an option"""


class TornadoNotAvailable(Exception):
    """use_tornado_loop does not work without Tornado dependency installed"""


class FuckFuckFuck(Exception):
    """test error"""
